from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from krita import *
import struct
import math

KI = Krita.instance()

class View3DViewport(QOpenGLWidget):
    _gl = None
    _tex = 0
    _texw = 0
    _texh = 0
    _dl = 0
    _verts = []
    _tgtx = 0
    _tgty = 0
    _tgtz = 0
    _rotx = 0
    _roty = 0
    _camd = 10
    _camRot = False
    _camPan = False
    _camZoom = False
    _viewFlat = False
    _viewFilter = False
    _viewWireframe = False
    _viewBackfaces = True
    _viewBgColor = QColor(26, 51, 76)
    _doc = None
    _mx = 0
    _my = 0
    _updateTPData = False
    _tiImage = None
    _tiImageOwnerDoc = None
    _tiImageDoc = None
    _fiLastLayer = None
    _periodicUpdates = True

    def __init__(self, parent=None):
        super(View3DViewport, self).__init__(parent)
        fmt = QSurfaceFormat()
        fmt.setRedBufferSize(8)
        fmt.setGreenBufferSize(8)
        fmt.setBlueBufferSize(8)
        fmt.setAlphaBufferSize(8)
        fmt.setSamples(0)
        fmt.setSwapBehavior(QSurfaceFormat.DoubleBuffer)
        fmt.setSwapInterval(0)
        fmt.setMajorVersion(2)
        fmt.setMinorVersion(1)
        fmt.setProfile(QSurfaceFormat.CompatibilityProfile)
        self.setFormat(fmt)
        self._timer = QTimer(self)
        self._timer.timeout.connect(self.timerTick)
        self._timer.start(33)
        self._toggleFilteringAction = QAction("Filtering")
        self._toggleFilteringAction.triggered.connect(self.toggleFiltering)
        self._toggleFilteringAction.setCheckable(True)
        self._toggleFilteringAction.setChecked(self._viewFilter)
        self._toggleShadingAction = QAction("Shading")
        self._toggleShadingAction.triggered.connect(self.toggleShading)
        self._toggleShadingAction.setCheckable(True)
        self._toggleShadingAction.setChecked(not self._viewFlat)
        self._toggleWireframeAction = QAction("Wireframe Overlay")
        self._toggleWireframeAction.triggered.connect(self.toggleWireframe)
        self._toggleWireframeAction.setCheckable(True)
        self._toggleWireframeAction.setChecked(self._viewWireframe)
        self._toggleBackfacesAction = QAction("Backfaces")
        self._toggleBackfacesAction.triggered.connect(self.toggleBackfaces)
        self._toggleBackfacesAction.setCheckable(True)
        self._toggleBackfacesAction.setChecked(self._viewBackfaces)
        self._resetViewAction = QAction("Reset View")
        self._resetViewAction.triggered.connect(self.resetView)
        self._backgroundColorAction = QAction("Background Color...")
        self._backgroundColorAction.triggered.connect(self.backgroundColor)
        self._toImageAction = QAction("Create Image from View")
        self._toImageAction.triggered.connect(self.toImage)
        self._fromImageAction = QAction("Backproject Last Created Image")
        self._fromImageAction.triggered.connect(self.fromImage)
        self._createMaskAction = QAction("Create Mask from View")
        self._createMaskAction.triggered.connect(self.createMask)
        self._createTexCoordMaskAction = QAction("Create Texture Coordinate Mask")
        self._createTexCoordMaskAction.triggered.connect(self.createTexCoordMask)
        self._createTexCoordWireframeAction = QAction("Create Texture Coordinate Wireframe")
        self._createTexCoordWireframeAction.triggered.connect(self.createTexCoordWireframe)
        self._setTextureToCurrentDocumentAction = QAction("Set Texture to Current Document")
        self._setTextureToCurrentDocumentAction.triggered.connect(self.setTextureToCurrentDocument)
        self._setTextureToAnyDocumentAction = QAction("Set Texture to Any Document")
        self._setTextureToAnyDocumentAction.triggered.connect(self.setTextureToAnyDocument)
        self._openMeshAction = QAction("Open Mesh...")
        self._openMeshAction.triggered.connect(self.openMesh)
        self._openRecentAction = [None]*10
        for i in range(10):
            self._openRecentAction[i] = QAction(KI.readSetting("view3dplugin", "recent" + str(i), ""))
            self._openRecentAction[i].triggered.connect(self.openRecent)
        self._clearRecentAction = QAction("Clear Recent Meshes")
        self._clearRecentAction.triggered.connect(self.clearRecent)

    def msgBox(self, msg):
        box = QMessageBox()
        box.setText(msg)
        box.setWindowTitle('View3D Message')
        box.setStandardButtons(QMessageBox.Ok)
        box.exec()

    def readJTF(self, fn):
        if self._dl != 0:
            self._gl.glDeleteLists(1, self._dl)
            self._dl = 0
        with open(fn, 'rb') as f:
            # Read header
            header = struct.unpack('<ccccLL', f.read(12))
            if header[0] != b'J' or header[1] != b'T' or header[2] != b'F' or header[3] != b'!':
                self.msgBox('Not a JTF file')
                return
            if header[4] != 0:
                self.msgBox('Unsupported JTF version')
            faces = header[5]
            # Read triangles
            self._verts = [None]*faces*3
            for i in range(faces*3):
                self._verts[i] = struct.unpack('<ffffffff', f.read(32))

    def getDoc(self):
        if self._doc != None:
            return self._doc
        return KI.activeDocument()

    def updateTexture(self):
        gl = self._gl
        if not gl:
            return
        if self._tex == 0:
           self._tex = gl.glGenTextures(1)
        gl.glBindTexture(gl.GL_TEXTURE_2D, self._tex)
        doc = self.getDoc()
        if doc:
            img = doc.projection(0, 0, doc.width(), doc.height())
            gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGBA, img.width(), img.height(), 0, gl.GL_BGRA, gl.GL_UNSIGNED_BYTE, img.bits().asarray(img.width()*img.height()*4))
            self._texw = doc.width()
            self._texh = doc.height()
        gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
        self.repaint()

    def timerTick(self):
        if self._periodicUpdates and self.isVisible():
            self.updateTexture()
        if self._doc != None or self._tiImageOwnerDoc != None:
            docFound = False
            iodFound = False
            for d in KI.documents():
                if d == self._doc:
                    docFound = True
                elif d == self._tiImageOwnerDoc:
                    iodFound = True
                if docFound and iodFound:
                    break
            if not docFound:
                self._doc = None
                self._fiLastLayer = None
            if not iodFound:
                self._tiImage = None
                self._tiImageDoc = None
                self._tiImageOwnerDoc = None

    def texCoordAtXY(self, x, y):
        idx = (self.height() - 1 - y) * self.width() + x
        return (int(round(-0.5 + ((int(self._tp[idx*4]*255.0) + int(self._tp[idx*4 + 1]*255)*256)/65535.0)*self._texw)), \
            int(round(-0.5 + ((int(self._tp[idx*4 + 2]*255.0) + int(self._tp[idx*4 + 3]*255)*256)/65535.0)*self._texh)))

    def plotPoint(self, x, y):
        doc = self.getDoc()
        if not doc:
            return
        x, y = self.texCoordAtXY(x, y)
        if x < 0 or y < 0 or x >= doc.width() or y >= doc.height():
            return
        doc.activeNode().setPixelData(QByteArray(b'\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF\xC0\x00\xC0\xFF'), x - 1, y - 1, 3, 3)
        doc.refreshProjection()

    def mouseMoveEvent(self, event):
        super(View3DViewport, self).mouseMoveEvent(event)
        if self._camRot:
            self._rotx += event.globalY() - self._my
            self._roty += event.globalX() - self._mx
            self._mx = event.globalX()
            self._my = event.globalY()
            self.repaint()
        if self._camPan:
            dx = -0.015*(event.globalX() - self._mx)
            dy = 0.015*(event.globalY() - self._my)
            rx = math.radians(self._rotx)
            ry = math.radians(self._roty)
            self._tgtx += math.cos(ry)*dx
            self._tgty += math.cos(rx)*dy
            self._tgtz += math.sin(ry)*dx - math.sin(rx)*dy
            self._mx = event.globalX()
            self._my = event.globalY()
            self.repaint()
        if self._camZoom:
            dy = 0.05*(event.globalY() - self._my)
            ry = math.radians(self._roty)
            self._camd += dy
            self._my = event.globalY()
            self.repaint()

    def mousePressEvent(self, event):
        super(View3DViewport, self).mousePressEvent(event)
        if self._tiImageDoc != None and self._tiImageDoc == KI.activeDocument():
            self._mx = event.globalX()
            self._my = event.globalY()
            return
        if event.button() == Qt.LeftButton:
            if event.modifiers() == Qt.AltModifier:
                self.plotPoint(event.x(), event.y())
                return
            self._camRot = True
            self._mx = event.globalX()
            self._my = event.globalY()
            return
        if event.button() == Qt.MiddleButton:
            self._camPan = True
            self._mx = event.globalX()
            self._my = event.globalY()
            return
        if event.button() == Qt.RightButton:
            self._camZoom = True
            self._mx = event.globalX()
            self._my = event.globalY()
            return

    def mouseReleaseEvent(self, event):
        super(View3DViewport, self).mouseReleaseEvent(event)
        self._updateTPData = True
        self.repaint()
        if event.button() == Qt.LeftButton:
            self._camRot = False
            return
        if event.button() == Qt.MiddleButton:
            self._camPan = False
            return
        if event.button() == Qt.RightButton:
            self._camZoom = False
            if self._mx == event.globalX() and self._my == event.globalY():
                self.showPopupMenu()
            return

    def showPopupMenu(self):
        menu = QMenu(self)
        if self._tiImageDoc != None and self._tiImageDoc == KI.activeDocument():
            menu.addAction(self._fromImageAction)
            menu.addSeparator()
        menu.addAction(self._toggleFilteringAction)
        menu.addAction(self._toggleShadingAction)
        menu.addAction(self._toggleWireframeAction)
        menu.addAction(self._toggleBackfacesAction)
        menu.addSeparator()
        menu.addAction(self._resetViewAction)
        menu.addAction(self._backgroundColorAction)
        menu.addSeparator()
        images = menu.addMenu("Images")
        images.addAction(self._toImageAction)
        images.addAction(self._fromImageAction)
        images.addSeparator()
        images.addAction(self._createMaskAction)
        images.addAction(self._createTexCoordMaskAction)
        images.addAction(self._createTexCoordWireframeAction)
        menu.addSeparator()
        texmenu = menu.addMenu("Texture")
        texmenu.addAction(self._setTextureToCurrentDocumentAction)
        texmenu.addAction(self._setTextureToAnyDocumentAction)
        menu.addSeparator()
        openmenu = menu.addMenu("Open")
        openmenu.addAction(self._openMeshAction)
        openmenu.addSeparator()
        anyRecent = False
        for i in range(10):
            if len(self._openRecentAction[i].text()) > 0:
                openmenu.addAction(self._openRecentAction[i])
                anyRecent = True
        if anyRecent:
            openmenu.addSeparator()
            openmenu.addAction(self._clearRecentAction)
        menu.popup(QPoint(self._mx, self._my))

    def resetView(self):
        self._tgtx = 0
        self._tgty = 0
        self._tgtz = 0
        self._rotx = 0
        self._roty = 0
        self._camd = 10
        self._updateTPData = True
        self.repaint()

    def backgroundColor(self):
        self._viewBgColor = QColorDialog.getColor(self._viewBgColor, self, "Background Color", QColorDialog.DontUseNativeDialog)
        self.repaint()

    def toggleShading(self):
        self._viewFlat = not self._viewFlat
        self.repaint()

    def toggleFiltering(self):
        self._viewFilter = not self._viewFilter
        self.repaint()

    def toggleWireframe(self):
        self._viewWireframe = not self._viewWireframe
        self.repaint()

    def toggleBackfaces(self):
        self._viewBackfaces = not self._viewBackfaces
        self.repaint()

    def saveRecents(self):
        for i in range(10):
            QAction(KI.writeSetting("view3dplugin", "recent" + str(i), self._openRecentAction[i].text()))

    def addToRecents(self, path):
        for i in range(10):
            if self._openRecentAction[i].text() == path:
                return
        for i in reversed(range(9)):
            self._openRecentAction[i + 1].setText(self._openRecentAction[i].text())
        self._openRecentAction[0].setText(path)
        self.saveRecents()

    def doOpenMesh(self, path):
        self._updateTPData = True
        self.readJTF(path)
        if KI.activeDocument() != None:
            self._doc = KI.activeDocument()

    def openMesh(self):
        fn = QFileDialog.getOpenFileName(self, 'Open Mesh', '', 'JTF files (*.jtf)')
        if fn and len(fn[0].strip()) > 0:
            fn = fn[0].strip()
            self.doOpenMesh(fn)
            self.addToRecents(fn)

    def openRecent(self):
        sender = self.sender()
        self.doOpenMesh(sender.text())

    def clearRecent(self):
        for i in range(10):
            self._openRecentAction[i].setText("")
        self.saveRecents()

    def toImage(self):
        gl = self._gl
        self.makeCurrent()
        gl.glClearColor(0.0, 0.0, 0.0, 0.0)
        self.paintGL()
        gl.glClearColor(self._viewBgColor.redF(), self._viewBgColor.greenF(), self._viewBgColor.blueF(), 0.0)
        rgbapixels = gl.glReadPixels(0, 0, self.width(), self.height(), gl.GL_BGRA, gl.GL_FLOAT)
        self.doneCurrent()
        data = bytearray()
        for rgba in rgbapixels:
            data.append(int(rgba*255.0))
        doc = KI.createDocument(self.width(), self.height(), 'Model', 'RGBA', 'U8', '', 100.0)
        self._tiImageOwnerDoc = doc
        bgrow = bytearray()
        for x in range(self.width()):
            bgrow.append(int(self._viewBgColor.blue()))
            bgrow.append(int(self._viewBgColor.green()))
            bgrow.append(int(self._viewBgColor.red()))
            bgrow.append(int(255))
        bgrow = QByteArray(bgrow)
        node = doc.rootNode().childNodes()[0]
        for y in range(self.height()):
            node.setPixelData(bgrow, 0, y, self.width(), 1)
        node = doc.createNode('Render', 'paintlayer')
        node.setOpacity(255)
        for y in range(self.height()):
            node.setPixelData(QByteArray(data[y*self.width()*4:]), 0, self.height() - 1 - y, self.width(), 1)
        doc.rootNode().addChildNode(node, None)
        node = doc.createGroupLayer('Layers to backproject')
        doc.rootNode().addChildNode(node, None)
        node.setOpacity(255)
        node.setPassThroughMode(False)
        node2 = doc.createNode('Paint Here', 'paintlayer')
        node2.setOpacity(255)
        node.addChildNode(node2, None)
        self._tiImage = node
        self._tiImageDoc = doc
        self._fiLastLayer = None
        KI.activeWindow().addView(doc)
        doc.refreshProjection()

    def fromImage(self):
        if self._tiImage == None:
            self.msgBox('Use "To Image" to create an image first')
            return
        doc = self.getDoc()
        if doc == None:
            self.msgBox('You need an active document')
            return
        rgba = bytearray(self._tiImage.projectionPixelData(0, 0, self.width(), self.height()))
        newrgba = bytearray(b'\x00'*(doc.width()*doc.height()*4))
        newrgbaf = [0.0]*(doc.width()*doc.height()*4)
        newrgbac = [0]*(doc.width()*doc.height())
        minrange = len(newrgba)
        maxrange = 0
        for y in range(self.height()):
            for x in range(self.width()):
                tc = self.texCoordAtXY(x, y)
                if tc[0] >= 0 and tc[0] < doc.width() and tc[1] >= 0 and tc[1] < doc.height():
                    srci = (y*self.width() + x)*4
                    tgti = (tc[1]*doc.width() + tc[0])*4
                    if rgba[srci + 3] > 0.0:
                        newrgbaf[tgti] += rgba[srci]
                        newrgbaf[tgti + 1] += rgba[srci + 1]
                        newrgbaf[tgti + 2] += rgba[srci + 2]
                        newrgbaf[tgti + 3] += rgba[srci + 3]
                        newrgbac[tc[1]*doc.width() + tc[0]] += 1
                        minrange = min(minrange, tc[1]*doc.width() + tc[0])
                        maxrange = max(maxrange, tc[1]*doc.width() + tc[0])
        for i in range(minrange, maxrange + 1):
            if newrgbac[i] > 0:
                newrgba[i*4] = int(newrgbaf[i*4]/newrgbac[i])
                newrgba[i*4 + 1] = int(newrgbaf[i*4 + 1]/newrgbac[i])
                newrgba[i*4 + 2] = int(newrgbaf[i*4 + 2]/newrgbac[i])
                newrgba[i*4 + 3] = int(newrgbaf[i*4 + 3]/newrgbac[i])
        if self._fiLastLayer == None:
            node = doc.createNode('Painted', 'paintlayer')
            node.setOpacity(255)
            doc.rootNode().addChildNode(node, None)
            self._fiLastLayer = node
        else:
            node = self._fiLastLayer
        node.setPixelData(QByteArray(newrgba), 0, 0, doc.width(), doc.height())
        doc.refreshProjection()

    def createMask(self):
        doc = self.getDoc()
        if doc == None:
            self.msgBox('You need an active document')
            return
        newrgba = bytearray(b'\x00'*(doc.width()*doc.height()*4))
        for y in range(self.height()):
            for x in range(self.width()):
                tc = self.texCoordAtXY(x, y)
                if tc[0] >= 0 and tc[0] < doc.width() and tc[1] >= 0 and tc[1] < doc.height():
                    tgti = (tc[1]*doc.width() + tc[0])*4
                    newrgba[tgti] = 255
                    newrgba[tgti + 1] = 255
                    newrgba[tgti + 2] = 255
                    newrgba[tgti + 3] = 255
        node = doc.createNode('View Mask', 'paintlayer')
        node.setOpacity(255)
        node.setPixelData(QByteArray(newrgba), 0, 0, doc.width(), doc.height())
        doc.rootNode().addChildNode(node, None)
        doc.refreshProjection()

    def createTexCoordLayer(self, solid):
        doc = self.getDoc()
        if doc == None:
            self.msgBox('You need an active document')
            return
        img = QImage(QSize(doc.width(), doc.height()), QImage.Format_ARGB32)
        img.fill(QColor(0, 0, 0, 0))
        poly = QPolygonF()
        poly.append(QPointF(0, 0))
        poly.append(QPointF(0, 0))
        poly.append(QPointF(0, 0))
        painter = QPainter()
        painter.begin(img)
        painter.setRenderHint(QPainter.Antialiasing, False)
        painter.setPen(Qt.white)
        if solid:
            painter.setBrush(Qt.white)
        for i in range(len(self._verts)//3):
            poly[0] = QPointF(self._verts[i*3][6]*doc.width(), self._verts[i*3][7]*doc.height())
            poly[1] = QPointF(self._verts[i*3 + 1][6]*doc.width(), self._verts[i*3 + 1][7]*doc.height())
            poly[2] = QPointF(self._verts[i*3 + 2][6]*doc.width(), self._verts[i*3 + 2][7]*doc.height())
            painter.drawPolygon(poly)
        painter.end()
        bits = img.bits()
        bits.setsize(img.byteCount())
        node = doc.createNode('Tex Coord Mask', 'paintlayer')
        node.setOpacity(255)
        node.setPixelData(QByteArray(bits.asstring()), 0, 0, doc.width(), doc.height())
        doc.rootNode().addChildNode(node, None)
        doc.refreshProjection()

    def createTexCoordMask(self):
        self.createTexCoordLayer(True)

    def createTexCoordWireframe(self):
        self.createTexCoordLayer(False)

    def setTextureToCurrentDocument(self):
        self._doc = KI.activeDocument()
        self.repaint()

    def setTextureToAnyDocument(self):
        self._doc = KI.activeDocument()
        self.repaint()

    def initializeGL(self):
        context = self.context()
        context.aboutToBeDestroyed.connect(self.cleanupGL)
        profile = QOpenGLVersionProfile()
        profile.setVersion(2, 1)
        profile.setProfile(QSurfaceFormat.NoProfile)
        gl = self._gl = context.versionFunctions(profile)
        self._tex = 0
        self.updateTexture()
        self._tprog = QOpenGLShaderProgram();
        self._tprog.addShaderFromSourceCode(QOpenGLShader.Vertex,
# Texture coord projection vertex shader
"""
#version 120
varying vec2 vTexCoord;
void main()
{
    vTexCoord = gl_MultiTexCoord0.xy;
    gl_Position = ftransform();
}
""")
        self._tprog.addShaderFromSourceCode(QOpenGLShader.Fragment,
# Texture coord projection fragment shader
"""
#version 120
#extension GL_EXT_gpu_shader4 : enable
varying vec2 vTexCoord;
float to01(float v) { v = fract(v); return v < 0.0 ? 1.0 - v : v; }
void main()
{
    gl_FragColor = vec4(
        (int(to01(vTexCoord.x)*65535.0) & 0xFF)/255.0,
        (int(to01(vTexCoord.x)*65535.0) >> 8)/255.0,
        (int(to01(vTexCoord.y)*65535.0) & 0xFF)/255.0,
        (int(to01(vTexCoord.y)*65535.0) >> 8)/255.0);
}
""")
        self._tprog.link()
        self._updateTPData = True
        gl.glClearColor(self._viewBgColor.redF(), self._viewBgColor.greenF(), self._viewBgColor.blueF(), 1.0)

    def cleanupGL(self):
        gl = self._gl
        self.context().makeCurrent()
        gl.glDeleteTextures(1, [self._tex])
        gl.glDeleteLists(1, self._dl)

    def doTheRender(self):
        gl = self._gl
        gl.glClearColor(self._viewBgColor.redF(), self._viewBgColor.greenF(), self._viewBgColor.blueF(), 1.0)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPushMatrix()
        f = 1.0/math.tan(math.radians(45*0.5))
        a = self.width()*1.0/self.height()
        gl.glLoadMatrixf([f/a, 0, 0, 0,
               0, f, 0, 0,
               0, 0, 10000.001/(0.001 - 10000.0), -1.0,
               0, 0, (2.0*10000.0*0.001)/(0.001 - 10000.0), 0])
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glPushMatrix()
        gl.glLoadIdentity()
        gl.glTranslatef(0, 0, -self._camd)
        gl.glRotatef(self._rotx, 1, 0, 0)
        gl.glRotatef(self._roty, 0, 1, 0)
        gl.glTranslatef(-self._tgtx, -self._tgty, -self._tgtz)
        gl.glColor3f(1, 1, 1)
        gl.glEnable(gl.GL_TEXTURE_2D)
        gl.glEnable(gl.GL_DEPTH_TEST)
        if self._viewFlat:
            gl.glDisable(gl.GL_LIGHTING)
        else:
            gl.glEnable(gl.GL_LIGHTING)
            gl.glEnable(gl.GL_LIGHT0)
        if self._viewBackfaces:
            gl.glDisable(gl.GL_CULL_FACE)
        else:
            gl.glEnable(gl.GL_CULL_FACE)
        gl.glBindTexture(gl.GL_TEXTURE_2D, self._tex)
        if self._viewFilter:
            gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR)
            gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR)
        else:
            gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR)
            gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_NEAREST)
        if self._dl == 0:
            self._dl = gl.glGenLists(1)
            gl.glNewList(self._dl, gl.GL_COMPILE)
            gl.glBegin(gl.GL_TRIANGLES)
            for v in self._verts:
                gl.glNormal3f(v[3], v[4], v[5])
                gl.glTexCoord2f(v[6], v[7])
                gl.glVertex3f(v[0], v[1], v[2])
            gl.glEnd()
            gl.glEndList()
        if self._updateTPData:
            gl.glClearColor(1.0, 1.0, 1.0, 1.0)
            gl.glClear(gl.GL_COLOR_BUFFER_BIT);
            self._updateTPData = False
            self._tprog.bind()
            gl.glCallList(self._dl)
            self._tprog.release()
            self._tp = gl.glReadPixels(0, 0, self.width(), self.height(), gl.GL_RGBA, gl.GL_FLOAT)
            gl.glClearColor(self._viewBgColor.redF(), self._viewBgColor.greenF(), self._viewBgColor.blueF(), 1.0)
            gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        gl.glCallList(self._dl)
        gl.glDisable(gl.GL_TEXTURE_2D)
        if self._viewWireframe:
            gl.glColor3f(1, 1, 1)
            gl.glDisable(gl.GL_LIGHTING)
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_LINE)
            gl.glEnable(gl.GL_POLYGON_OFFSET_LINE)
            gl.glPolygonOffset(self._camd*0.001, -1)
            gl.glCallList(self._dl)
            gl.glDisable(gl.GL_POLYGON_OFFSET_LINE)
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)
        gl.glPopMatrix()
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glPopMatrix()
        gl.glMatrixMode(gl.GL_MODELVIEW)

    def paintGL(self):
        self.doTheRender()

    def resizeGL(self, width, height):
        gl = self._gl
        gl.glViewport(0, 0, width, height)


class View3DDocker(DockWidget):

    def __init__(self):
        super().__init__()
        self.setWindowTitle("View 3D")
        self._viewport = View3DViewport()

        layout = QVBoxLayout()
        layout.addWidget(self._viewport)

        buttons = QHBoxLayout()

        chk = QCheckBox("Live")
        chk.setCheckState(Qt.Checked)
        chk.stateChanged.connect(self.liveChanged)
        buttons.addWidget(chk)

        layout.addLayout(buttons)

        w = QWidget(self)
        w.setLayout(layout)
        self.setWidget(w)


    def liveChanged(self, state):
        self._viewport._periodicUpdates = state != Qt.Unchecked
        pass

    def canvasChanged(self, canvas):
        if self._viewport._gl:
            self._viewport.updateTexture()
        pass

Krita.instance().addDockWidgetFactory(DockWidgetFactory("view3d", DockWidgetFactoryBase.DockRight, View3DDocker))
